#include <stdio.h>
#include <stdlib.h>

#ifdef USE_NLLIBC
#include <nllibc.h>
#endif

#include "lib.h"
#include "type.h"
#include "asm_symbol.h"
#include "asm_code.h"
#include "asm_code_amd64.h"

static int asm_code_amd64_get_word_size(void)
{
  return 8;
}

static int asm_code_amd64_get_pointer_size(void)
{
  return asm_code_amd64_get_word_size();
}

static int asm_code_amd64_funccall_args_reg_number(void)
{
  return 6; /* RDI/RSI/RDX/RCX/R8/R9 (See funccall_regs[]) */
}

static int asm_code_amd64_funccall_args_stack_number(void)
{
  return VALUE_FUNCTION_ARGS_NUM;
}

static int asm_code_amd64_tmp_reg_number(void)
{
  return 6; /* R10/R11/R12/R13/R14/R15 (See tmp_regs[]) */
}

static int asm_code_amd64_tmp_stack_number(void)
{
  return VALUE_FUNCTION_ARGS_NUM;
}

static int asm_code_amd64_function_register_number(void)
{
  return 5; /* RBX/R12/R13/R14/R15 (See asm_code_amd64_function_register_save() */
}

static int asm_code_amd64_function_saveparam_number(void)
{
  return 2; /* RBP and return address (See asm_code_amd64_function_start() */
}

static int asm_code_amd64_stack_align_size(void)
{
  return 16;
}

static int asm_code_amd64_stack_correct_size(void)
{
  return 0;
}

static char *r0_by_size(model_t model)
{
  if (model != NULL) {
    switch (model_get_size(model)) {
    case -1: ERREXIT(1);
    case  1: return "al";
    case  2: return "ax";
    case  4: return "eax";
    default: break;
    }
  }
  return "rax";
}

static char *r1_by_size(model_t model)
{
  if (model != NULL) {
    switch (model_get_size(model)) {
    case -1: ERREXIT(1);
    case  1: return "bl";
    case  2: return "bx";
    case  4: return "ebx";
    default: break;
    }
  }
  return "rbx";
}

static void asm_code_amd64_get_value(FILE *out, long value)
{
  ASM_CODE_OUT(out, "\tmov\t$0x%lx, %%rax\n", value);
}

static void asm_code_amd64_get_value_r1(FILE *out, long value)
{
  ASM_CODE_OUT(out, "\tmov\t$0x%lx, %%rbx\n", value);
}

static void asm_code_amd64_get_address_stack(FILE *out)
{
  ASM_CODE_OUT(out, "\tmov\t%%rsp, %%rax\n");
}

static void asm_code_amd64_get_address(FILE *out, char *label)
{
  ASM_CODE_OUT(out, "\tmov\t$%s, %%rax\n", label);
}

static void asm_code_amd64_add_address(FILE *out, int offset)
{
  ASM_CODE_OUT(out, "\tlea\t0x%x(%%rax), %%rax\n", offset);
}

static void asm_code_amd64_get_r1(FILE *out)
{
  ASM_CODE_OUT(out, "\tmov\t%%rbx, %%rax\n");
}

static void asm_code_amd64_set_r1(FILE *out)
{
  ASM_CODE_OUT(out, "\tmov\t%%rax, %%rbx\n");
}

static void asm_code_amd64_memory_load(FILE *out, int offset, model_t model)
{
  ASM_CODE_OUT(out, "\tmov\t0x%x(%%rbx), %%%s\n", offset, r0_by_size(model));
}

static void asm_code_amd64_memory_store(FILE *out, int offset, model_t model)
{
  ASM_CODE_OUT(out, "\tmov\t%%%s, 0x%x(%%rbx)\n", r0_by_size(model), offset);
}

static void asm_code_amd64_stack_load(FILE *out, int offset, model_t model)
{
  ASM_CODE_OUT(out, "\tmov\t0x%x(%%rsp), %%%s\n", offset, r0_by_size(model));
}

static void asm_code_amd64_stack_store(FILE *out, int offset, model_t model)
{
  ASM_CODE_OUT(out, "\tmov\t%%%s, 0x%x(%%rsp)\n", r0_by_size(model), offset);
}

static void asm_code_amd64_stack_load_r1(FILE *out, int offset, model_t model)
{
  ASM_CODE_OUT(out, "\tmov\t0x%x(%%rsp), %%%s\n", offset, r1_by_size(model));
}

static void asm_code_amd64_stack_store_r1(FILE *out, int offset, model_t model)
{
  ASM_CODE_OUT(out, "\tmov\t%%%s, 0x%x(%%rsp)\n", r1_by_size(model), offset);
}

static void asm_code_amd64_stack_expand(FILE *out, int size)
{
  ASM_CODE_OUT(out, "\tsub\t$0x%x, %%rsp\n", size);
}

static void asm_code_amd64_stack_reduce(FILE *out, int size)
{
  ASM_CODE_OUT(out, "\tadd\t$0x%x, %%rsp\n", size);
}

static char *funccall_regs[] = {
  "rdi", "rsi", "rdx", "rcx", "r8", "r9"
};

static void asm_code_amd64_funccall_reg_load(FILE *out, int number)
{
  ASM_CODE_OUT(out, "\tmov\t%%%s, %%rax\n", funccall_regs[number]);
}

static void asm_code_amd64_funccall_reg_store(FILE *out, int number)
{
  ASM_CODE_OUT(out, "\tmov\t%%rax, %%%s\n", funccall_regs[number]);
}

static char *tmp_regs[] = {
  "r10", "r11", "r12", "r13", "r14", "r15",
};

static void asm_code_amd64_tmp_reg_load(FILE *out, int number)
{
  ASM_CODE_OUT(out, "\tmov\t%%%s, %%rax\n", tmp_regs[number]);
}

static void asm_code_amd64_tmp_reg_save(FILE *out, int number)
{
  ASM_CODE_OUT(out, "\tmov\t%%rax, %%%s\n", tmp_regs[number]);
}

static void asm_code_amd64_tmp_reg_load_r1(FILE *out, int number)
{
  ASM_CODE_OUT(out, "\tmov\t%%%s, %%rbx\n", tmp_regs[number]);
}

static void asm_code_amd64_tmp_reg_save_r1(FILE *out, int number)
{
  ASM_CODE_OUT(out, "\tmov\t%%rbx, %%%s\n", tmp_regs[number]);
}

static int asm_code_amd64_sign_extension_char(FILE *out)
{
  ASM_CODE_OUT(out, "\tmovsbq\t%%al, %%rax\n");
  return 0;
}

static int asm_code_amd64_sign_extension_uchar(FILE *out)
{
  ASM_CODE_OUT(out, "\tmovzbl\t%%al, %%eax\n");
  return 0;
}

static int asm_code_amd64_sign_extension_short(FILE *out)
{
  ASM_CODE_OUT(out, "\tmovswq\t%%ax, %%rax\n");
  return 0;
}

static int asm_code_amd64_sign_extension_ushort(FILE *out)
{
  ASM_CODE_OUT(out, "\tmovzwl\t%%ax, %%eax\n");
  return 0;
}

static int asm_code_amd64_sign_extension_int(FILE *out)
{
  ASM_CODE_OUT(out, "\tmovslq\t%%eax, %%rax\n");
  return 0;
}

static int asm_code_amd64_sign_extension_uint(FILE *out)
{
  ASM_CODE_OUT(out, "\tmov\t%%eax, %%eax\n");
  return 0;
}

static int asm_code_amd64_calc_inv(FILE *out, model_t model, model_t model_arg0)
{
  ASM_CODE_OUT(out, "\tnot\t%%rax\n");
  return 0;
}

static int asm_code_amd64_calc_minus(FILE *out, model_t model, model_t model_arg0)
{
  ASM_CODE_OUT(out, "\tneg\t%%rax\n");
  return 0;
}

static int asm_code_amd64_calc_op1(FILE *out, c_type_t type, model_t model, model_t model_arg0)
{
  return -1;
}

static void asm_code_amd64_calc_add(FILE *out, model_t model, model_t model_arg0, model_t model_arg1)
{
  ASM_CODE_OUT(out, "\tadd\t%%rbx, %%rax\n");
}

static void asm_code_amd64_calc_sub(FILE *out, model_t model, model_t model_arg0, model_t model_arg1)
{
  ASM_CODE_OUT(out, "\tsub\t%%rbx, %%rax\n");
}

static void asm_code_amd64_calc_and(FILE *out, model_t model, model_t model_arg0, model_t model_arg1)
{
  ASM_CODE_OUT(out, "\tand\t%%rbx, %%rax\n");
}

static void asm_code_amd64_calc_or(FILE *out, model_t model, model_t model_arg0, model_t model_arg1)
{
  ASM_CODE_OUT(out, "\tor\t%%rbx, %%rax\n");
}

static void asm_code_amd64_calc_xor(FILE *out, model_t model, model_t model_arg0, model_t model_arg1)
{
  ASM_CODE_OUT(out, "\txor\t%%rbx, %%rax\n");
}

static int asm_code_amd64_calc_mul(FILE *out, model_t model, model_t model_arg0, model_t model_arg1)
{
  ASM_CODE_OUT(out, "\timul\t%%%s\n", r1_by_size(model_arg0));
  return 0;
}

static int asm_code_amd64_calc_div(FILE *out, model_t model, model_t model_arg0, model_t model_arg1)
{
  ASM_CODE_OUT(out, "\tmov\t%%rbx, %%rcx\n");
  if (model_arg0->flags & MODEL_FLAG_UNSIGNED) {
    ASM_CODE_OUT(out, "\tmov\t$0, %%rdx\n");
  } else {
    ASM_CODE_OUT(out, "\tcqto\n");
  }
  if (model->flags & MODEL_FLAG_UNSIGNED) {
    switch (model_get_size(model_arg1)) {
    case -1: ERREXIT(1);
    case  1: ASM_CODE_OUT(out, "\tdiv\t%%cl\n"); break;
    case  2: ASM_CODE_OUT(out, "\tdiv\t%%cx\n"); break;
    case  4: ASM_CODE_OUT(out, "\tdiv\t%%ecx\n"); break;
    default: ASM_CODE_OUT(out, "\tdiv\t%%rcx\n"); break;
    }
  } else {
    ASM_CODE_OUT(out, "\tidiv\t%%rcx\n");
  }
  return 0;
}

static int asm_code_amd64_calc_mod(FILE *out, model_t model, model_t model_arg0, model_t model_arg1)
{
  ASM_CODE_OUT(out, "\tmov\t%%rbx, %%rcx\n");
  if (model_arg0->flags & MODEL_FLAG_UNSIGNED) {
    ASM_CODE_OUT(out, "\tmov\t$0, %%rdx\n");
  } else {
    ASM_CODE_OUT(out, "\tcqto\n");
  }
  if (model->flags & MODEL_FLAG_UNSIGNED) {
    ASM_CODE_OUT(out, "\tdiv\t%%rcx\n");
  } else {
    ASM_CODE_OUT(out, "\tidiv\t%%rcx\n");
  }
  ASM_CODE_OUT(out, "\tmov\t%%rdx, %%rax\n");
  return 0;
}

static int asm_code_amd64_calc_llshift(FILE *out, model_t model, model_t model_arg0, model_t model_arg1)
{
  ASM_CODE_OUT(out, "\tmov\t%%rbx, %%rcx\n");
  ASM_CODE_OUT(out, "\tshl\t%%cl, %%%s\n", r0_by_size(model_arg0));
  return 0;
}

static int asm_code_amd64_calc_rashift(FILE *out, model_t model, model_t model_arg0, model_t model_arg1)
{
  ASM_CODE_OUT(out, "\tmov\t%%rbx, %%rcx\n");
  ASM_CODE_OUT(out, "\tsar\t%%cl, %%%s\n", r0_by_size(model_arg0));
  return 0;
}

static int asm_code_amd64_calc_rlshift(FILE *out, model_t model, model_t model_arg0, model_t model_arg1)
{
  ASM_CODE_OUT(out, "\tmov\t%%rbx, %%rcx\n");
  ASM_CODE_OUT(out, "\tshr\t%%cl, %%%s\n", r0_by_size(model_arg0));
  return 0;
}

static int asm_code_amd64_calc_op2(FILE *out, c_type_t type, model_t model, model_t model_arg0, model_t model_arg1)
{
  return -1;
}

static void asm_code_amd64_branch(FILE *out, char *label)
{
  ASM_CODE_OUT(out, "\tjmp\t%s\n", label);
}

static int asm_code_amd64_branch_zero(FILE *out, char *label)
{
  ASM_CODE_OUT(out, "\tcmp\t$0, %%rax\n");
  ASM_CODE_OUT(out, "\tje\t%s\n", label);
  return 0;
}

static int asm_code_amd64_branch_nzero(FILE *out, char *label)
{
  ASM_CODE_OUT(out, "\tcmp\t$0, %%rax\n");
  ASM_CODE_OUT(out, "\tjne\t%s\n", label);
  return 0;
}

static int asm_code_amd64_branch_cmp_eq(FILE *out, char *label, model_t model)
{
  ASM_CODE_OUT(out, "\tcmp\t%%rbx, %%rax\n");
  ASM_CODE_OUT(out, "\tje\t%s\n", label);
  return 0;
}

static int asm_code_amd64_branch_cmp_ne(FILE *out, char *label, model_t model)
{
  ASM_CODE_OUT(out, "\tcmp\t%%rbx, %%rax\n");
  ASM_CODE_OUT(out, "\tjne\t%s\n", label);
  return 0;
}

static int asm_code_amd64_branch_cmp_lt(FILE *out, char *label, model_t model)
{
  ASM_CODE_OUT(out, "\tcmp\t%%rbx, %%rax\n");
  ASM_CODE_OUT(out, "\tjl\t%s\n", label);
  return 0;
}

static int asm_code_amd64_branch_cmp_gt(FILE *out, char *label, model_t model)
{
  ASM_CODE_OUT(out, "\tcmp\t%%rbx, %%rax\n");
  ASM_CODE_OUT(out, "\tjg\t%s\n", label);
  return 0;
}

static int asm_code_amd64_branch_cmp_le(FILE *out, char *label, model_t model)
{
  ASM_CODE_OUT(out, "\tcmp\t%%rbx, %%rax\n");
  ASM_CODE_OUT(out, "\tjle\t%s\n", label);
  return 0;
}

static int asm_code_amd64_branch_cmp_ge(FILE *out, char *label, model_t model)
{
  ASM_CODE_OUT(out, "\tcmp\t%%rbx, %%rax\n");
  ASM_CODE_OUT(out, "\tjge\t%s\n", label);
  return 0;
}

static int asm_code_amd64_branch_cmp_ult(FILE *out, char *label, model_t model)
{
  ASM_CODE_OUT(out, "\tcmp\t%%rbx, %%rax\n");
  ASM_CODE_OUT(out, "\tjb\t%s\n", label);
  return 0;
}

static int asm_code_amd64_branch_cmp_ugt(FILE *out, char *label, model_t model)
{
  ASM_CODE_OUT(out, "\tcmp\t%%rbx, %%rax\n");
  ASM_CODE_OUT(out, "\tja\t%s\n", label);
  return 0;
}

static int asm_code_amd64_branch_cmp_ule(FILE *out, char *label, model_t model)
{
  ASM_CODE_OUT(out, "\tcmp\t%%rbx, %%rax\n");
  ASM_CODE_OUT(out, "\tjbe\t%s\n", label);
  return 0;
}

static int asm_code_amd64_branch_cmp_uge(FILE *out, char *label, model_t model)
{
  ASM_CODE_OUT(out, "\tcmp\t%%rbx, %%rax\n");
  ASM_CODE_OUT(out, "\tjae\t%s\n", label);
  return 0;
}

static int asm_code_amd64_branch_cmp(FILE *out, char *label, c_type_t type, model_t model)
{
  return -1;
}

static void asm_code_amd64_function_call(FILE *out, char *label)
{
  ASM_CODE_OUT(out, "\tcallq\t%s\n", label);
}

static void asm_code_amd64_function_call_pointer(FILE *out)
{
  ASM_CODE_OUT(out, "\tcallq\t*%%rax\n");
}

static void asm_code_amd64_function_start(FILE *out)
{
  ASM_CODE_OUT(out, "\tpushq\t%%rbp\n");
  ASM_CODE_OUT(out, "\tmov\t%%rsp, %%rbp\n");
}

static void asm_code_amd64_function_end(FILE *out)
{
  ASM_CODE_OUT(out, "\tleaveq\n");
  ASM_CODE_OUT(out, "\tretq\n\n");
}

static void asm_code_amd64_function_register_save(FILE *out)
{
  ASM_CODE_OUT(out, "\tpushq\t%%rbx\n");
  ASM_CODE_OUT(out, "\tpushq\t%%r12\n");
  ASM_CODE_OUT(out, "\tpushq\t%%r13\n");
  ASM_CODE_OUT(out, "\tpushq\t%%r14\n");
  ASM_CODE_OUT(out, "\tpushq\t%%r15\n");
}

static void asm_code_amd64_function_register_load(FILE *out)
{
  ASM_CODE_OUT(out, "\tpopq\t%%r15\n");
  ASM_CODE_OUT(out, "\tpopq\t%%r14\n");
  ASM_CODE_OUT(out, "\tpopq\t%%r13\n");
  ASM_CODE_OUT(out, "\tpopq\t%%r12\n");
  ASM_CODE_OUT(out, "\tpopq\t%%rbx\n");
}

struct asm_code_callback asm_code_callback_amd64 = {
  asm_code_amd64_get_word_size,
  asm_code_amd64_get_pointer_size,

  asm_code_amd64_funccall_args_reg_number,
  asm_code_amd64_funccall_args_stack_number,
  asm_code_amd64_tmp_reg_number,
  asm_code_amd64_tmp_stack_number,
  asm_code_amd64_function_register_number,
  asm_code_amd64_function_saveparam_number,

  asm_code_amd64_stack_align_size,
  asm_code_amd64_stack_correct_size,

  asm_code_amd64_get_value,
  asm_code_amd64_get_value_r1,
  asm_code_amd64_get_address_stack,
  asm_code_amd64_get_address,
  asm_code_amd64_add_address,
  asm_code_amd64_get_r1,
  asm_code_amd64_set_r1,

  asm_code_amd64_memory_load,
  asm_code_amd64_memory_store,
  asm_code_amd64_stack_load,
  asm_code_amd64_stack_store,
  asm_code_amd64_stack_load_r1,
  asm_code_amd64_stack_store_r1,
  asm_code_amd64_stack_expand,
  asm_code_amd64_stack_reduce,

  asm_code_amd64_funccall_reg_load,
  asm_code_amd64_funccall_reg_store,
  asm_code_amd64_tmp_reg_load,
  asm_code_amd64_tmp_reg_save,
  asm_code_amd64_tmp_reg_load_r1,
  asm_code_amd64_tmp_reg_save_r1,

  asm_code_amd64_sign_extension_char,
  asm_code_amd64_sign_extension_uchar,
  asm_code_amd64_sign_extension_short,
  asm_code_amd64_sign_extension_ushort,
  asm_code_amd64_sign_extension_int,
  asm_code_amd64_sign_extension_uint,

  asm_code_amd64_calc_inv,
  asm_code_amd64_calc_minus,

  asm_code_amd64_calc_op1,

  asm_code_amd64_calc_add,
  asm_code_amd64_calc_sub,
  asm_code_amd64_calc_and,
  asm_code_amd64_calc_or,
  asm_code_amd64_calc_xor,

  asm_code_amd64_calc_mul,
  asm_code_amd64_calc_div,
  asm_code_amd64_calc_mod,
  asm_code_amd64_calc_llshift,
  asm_code_amd64_calc_rashift,
  asm_code_amd64_calc_rlshift,

  asm_code_amd64_calc_op2,

  asm_code_amd64_branch,
  asm_code_amd64_branch_zero,
  asm_code_amd64_branch_nzero,

  asm_code_amd64_branch_cmp_eq,
  asm_code_amd64_branch_cmp_ne,

  asm_code_amd64_branch_cmp_lt,
  asm_code_amd64_branch_cmp_gt,
  asm_code_amd64_branch_cmp_le,
  asm_code_amd64_branch_cmp_ge,
  asm_code_amd64_branch_cmp_ult,
  asm_code_amd64_branch_cmp_ugt,
  asm_code_amd64_branch_cmp_ule,
  asm_code_amd64_branch_cmp_uge,

  asm_code_amd64_branch_cmp,

  asm_code_amd64_function_call,
  asm_code_amd64_function_call_pointer,

  asm_code_amd64_function_start,
  asm_code_amd64_function_end,
  asm_code_amd64_function_register_save,
  asm_code_amd64_function_register_load,
};
