#include <stdio.h>
#include <stdlib.h>

#ifdef USE_NLLIBC
#include <nllibc.h>
#endif

#include "lib.h"
#include "type.h"
#include "asm_symbol.h"
#include "asm_format.h"
#include "asm_code.h"
#include "asm_code_thumb.h"

static int asm_code_thumb_get_word_size(void)
{
  return 4;
}

static int asm_code_thumb_get_pointer_size(void)
{
  return asm_code_thumb_get_word_size();
}

static int asm_code_thumb_funccall_args_reg_number(void)
{
  return 4; /* r0/r1/r2/r3 (See funccall_regs[]) */
}

static int asm_code_thumb_funccall_args_stack_number(void)
{
  return VALUE_FUNCTION_ARGS_NUM;
}

static int asm_code_thumb_tmp_reg_number(void)
{
  return 6; /* r0/r1/r2/r3/r4/r5 (See tmp_regs[]) */
}

static int asm_code_thumb_tmp_stack_number(void)
{
  return VALUE_FUNCTION_ARGS_NUM;
}

static int asm_code_thumb_function_register_number(void)
{
  return 2; /* r4/r5 (See asm_code_thumb_function_register_save() */
}

static int asm_code_thumb_function_saveparam_number(void)
{
  return 1; /* lr (See asm_code_thumb_function_start() */
}

static int asm_code_thumb_stack_align_size(void)
{
  return 16;
}

static int asm_code_thumb_stack_correct_size(void)
{
  return 0;
}

static char *load_by_size(model_t model)
{
  if (model != NULL) {
    switch (model_get_size(model)) {
    case -1: ERREXIT(1);
    case  1: return "ldrb";
    case  2: return "ldrh";
    default: break;
    }
  }
  return "ldr";
}

static char *store_by_size(model_t model)
{
  if (model != NULL) {
    switch (model_get_size(model)) {
    case -1: ERREXIT(1);
    case  1: return "strb";
    case  2: return "strh";
    default: break;
    }
  }
  return "str";
}

static void asm_code_thumb_get_value(FILE *out, long value)
{
  asm_symbol_tmp_label_reserve(1);
  ASM_CODE_OUT(out, "\tldr\tr0, [pc, #0]\n");
  ASM_CODE_OUT(out, "\tb\t%s\n", asm_symbol_tmp_label_namef(0));
  ASM_CODE_OUT(out, "\t.word\t0x%lx\n", value);
  asm_format_label(out, asm_symbol_tmp_label_name(0));
  asm_symbol_tmp_label_release(1);
}

static void asm_code_thumb_get_value_r1(FILE *out, long value)
{
  asm_symbol_tmp_label_reserve(1);
  ASM_CODE_OUT(out, "\tldr\tr1, [pc, #0]\n");
  ASM_CODE_OUT(out, "\tb\t%s\n", asm_symbol_tmp_label_namef(0));
  ASM_CODE_OUT(out, "\t.word\t0x%lx\n", value);
  asm_format_label(out, asm_symbol_tmp_label_name(0));
  asm_symbol_tmp_label_release(1);
}

static void asm_code_thumb_get_address_stack(FILE *out)
{
  ASM_CODE_OUT(out, "\tmov\tr0, sp\n");
}

static void asm_code_thumb_get_address(FILE *out, char *label)
{
  asm_symbol_tmp_label_reserve(1);
  ASM_CODE_OUT(out, "\tldr\tr0, [pc, #0]\n");
  ASM_CODE_OUT(out, "\tb\t%s\n", asm_symbol_tmp_label_namef(0));
  ASM_CODE_OUT(out, "\t.word\t%s\n", label);
  asm_format_label(out, asm_symbol_tmp_label_name(0));
  asm_symbol_tmp_label_release(1);
}

static void asm_code_thumb_add_address(FILE *out, int offset)
{
  ASM_CODE_OUT(out, "\tadds\tr0, #0x%x\n", offset);
}

static void asm_code_thumb_get_r1(FILE *out)
{
  ASM_CODE_OUT(out, "\tmov\tr0, r1\n");
}

static void asm_code_thumb_set_r1(FILE *out)
{
  ASM_CODE_OUT(out, "\tmov\tr1, r0\n");
}

static void asm_code_thumb_memory_load(FILE *out, int offset, model_t model)
{
  ASM_CODE_OUT(out, "\t%s\tr0, [r1, #0x%x]\n", load_by_size(model), offset);
}

static void asm_code_thumb_memory_store(FILE *out, int offset, model_t model)
{
  ASM_CODE_OUT(out, "\t%s\tr0, [r1, #0x%x]\n", store_by_size(model), offset);
}

static void asm_code_thumb_stack_load(FILE *out, int offset, model_t model)
{
  ASM_CODE_OUT(out, "\t%s\tr0, [sp, #0x%x]\n", load_by_size(model), offset);
}

static void asm_code_thumb_stack_store(FILE *out, int offset, model_t model)
{
  ASM_CODE_OUT(out, "\t%s\tr0, [sp, #0x%x]\n", store_by_size(model), offset);
}

static void asm_code_thumb_stack_load_r1(FILE *out, int offset, model_t model)
{
  ASM_CODE_OUT(out, "\t%s\tr1, [sp, #0x%x]\n", load_by_size(model), offset);
}

static void asm_code_thumb_stack_store_r1(FILE *out, int offset, model_t model)
{
  ASM_CODE_OUT(out, "\t%s\tr1, [sp, #0x%x]\n", store_by_size(model), offset);
}

static void asm_code_thumb_stack_expand(FILE *out, int size)
{
  ASM_CODE_OUT(out, "\tsub\tsp, #0x%x\n", size);
}

static void asm_code_thumb_stack_reduce(FILE *out, int size)
{
  ASM_CODE_OUT(out, "\tadd\tsp, #0x%x\n", size);
}

static char *funccall_regs[] = {
  "r0", "r1", "r2", "r3"
};

static void asm_code_thumb_funccall_reg_load(FILE *out, int number)
{
  ASM_CODE_OUT(out, "\tmov\tr0, %s\n", funccall_regs[number]);
}

static void asm_code_thumb_funccall_reg_store(FILE *out, int number)
{
  ASM_CODE_OUT(out, "\tmov\t%s, r0\n", funccall_regs[number]);
}

static char *tmp_regs[] = {
  "r0", "r1", "r2", "r3", "r4", "r5",
};

static void asm_code_thumb_tmp_reg_load(FILE *out, int number)
{
  ASM_CODE_OUT(out, "\tmov\tr0, %s\n", tmp_regs[number]);
}

static void asm_code_thumb_tmp_reg_save(FILE *out, int number)
{
  ASM_CODE_OUT(out, "\tmov\t%s, r0\n", tmp_regs[number]);
}

static void asm_code_thumb_tmp_reg_load_r1(FILE *out, int number)
{
  ASM_CODE_OUT(out, "\tmov\tr1, %s\n", tmp_regs[number]);
}

static void asm_code_thumb_tmp_reg_save_r1(FILE *out, int number)
{
  ASM_CODE_OUT(out, "\tmov\t%s, r1\n", tmp_regs[number]);
}

static int asm_code_thumb_sign_extension_char(FILE *out)
{
  ASM_CODE_OUT(out, "\tlsls\tr0, r0, #24\n");
  ASM_CODE_OUT(out, "\tasrs\tr0, r0, #24\n");
  return 0;
}

static int asm_code_thumb_sign_extension_uchar(FILE *out)
{
  ASM_CODE_OUT(out, "\tlsls\tr0, r0, #24\n");
  ASM_CODE_OUT(out, "\tlsrs\tr0, r0, #24\n");
  return 0;
}

static int asm_code_thumb_sign_extension_short(FILE *out)
{
  ASM_CODE_OUT(out, "\tlsls\tr0, r0, #16\n");
  ASM_CODE_OUT(out, "\tasrs\tr0, r0, #16\n");
  return 0;
}

static int asm_code_thumb_sign_extension_ushort(FILE *out)
{
  ASM_CODE_OUT(out, "\tlsls\tr0, r0, #16\n");
  ASM_CODE_OUT(out, "\tlsrs\tr0, r0, #16\n");
  return 0;
}

static int asm_code_thumb_sign_extension_int(FILE *out)
{
  return -1;
}

static int asm_code_thumb_sign_extension_uint(FILE *out)
{
  return -1;
}

static int asm_code_thumb_calc_inv(FILE *out, model_t model, model_t model_arg0)
{
  ASM_CODE_OUT(out, "\tmvns\tr0, r0\n");
  return 0;
}

static int asm_code_thumb_calc_minus(FILE *out, model_t model, model_t model_arg0)
{
  ASM_CODE_OUT(out, "\tnegs\tr0, r0\n");
  return 0;
}

static int asm_code_thumb_calc_op1(FILE *out, c_type_t type, model_t model, model_t model_arg0)
{
  return -1;
}

static void asm_code_thumb_calc_add(FILE *out, model_t model, model_t model_arg0, model_t model_arg1)
{
  ASM_CODE_OUT(out, "\tadds\tr0, r0, r1\n");
}

static void asm_code_thumb_calc_sub(FILE *out, model_t model, model_t model_arg0, model_t model_arg1)
{
  ASM_CODE_OUT(out, "\tsubs\tr0, r0, r1\n");
}

static void asm_code_thumb_calc_and(FILE *out, model_t model, model_t model_arg0, model_t model_arg1)
{
  ASM_CODE_OUT(out, "\tands\tr0, r1\n");
}

static void asm_code_thumb_calc_or(FILE *out, model_t model, model_t model_arg0, model_t model_arg1)
{
  ASM_CODE_OUT(out, "\torrs\tr0, r1\n");
}

static void asm_code_thumb_calc_xor(FILE *out, model_t model, model_t model_arg0, model_t model_arg1)
{
  ASM_CODE_OUT(out, "\teors\tr0, r1\n");
}

static int asm_code_thumb_calc_mul(FILE *out, model_t model, model_t model_arg0, model_t model_arg1)
{
  ASM_CODE_OUT(out, "\tmuls\tr0, r1\n");
  return 0;
}

static int asm_code_thumb_calc_div(FILE *out, model_t model, model_t model_arg0, model_t model_arg1)
{
  return -1; /* use builtin library */
}

static int asm_code_thumb_calc_mod(FILE *out, model_t model, model_t model_arg0, model_t model_arg1)
{
  return -1; /* use builtin library */
}

static int asm_code_thumb_calc_llshift(FILE *out, model_t model, model_t model_arg0, model_t model_arg1)
{
  ASM_CODE_OUT(out, "\tlsls\tr0, r1\n");
  return 0;
}

static int asm_code_thumb_calc_rashift(FILE *out, model_t model, model_t model_arg0, model_t model_arg1)
{
  ASM_CODE_OUT(out, "\tasrs\tr0, r1\n");
  return 0;
}

static int asm_code_thumb_calc_rlshift(FILE *out, model_t model, model_t model_arg0, model_t model_arg1)
{
  ASM_CODE_OUT(out, "\tlsrs\tr0, r1\n");
  return 0;
}

static int asm_code_thumb_calc_op2(FILE *out, c_type_t type, model_t model, model_t model_arg0, model_t model_arg1)
{
  return -1;
}

static void asm_code_thumb_branch(FILE *out, char *label)
{
  ASM_CODE_OUT(out, "\tb\t%s\n", label);
}

static int asm_code_thumb_branch_zero(FILE *out, char *label)
{
  ASM_CODE_OUT(out, "\tcmp\tr0, #0\n");
  ASM_CODE_OUT(out, "\tbeq.n\t%s\n", label);
  return 0;
}

static int asm_code_thumb_branch_nzero(FILE *out, char *label)
{
  ASM_CODE_OUT(out, "\tcmp\tr0, #0\n");
  ASM_CODE_OUT(out, "\tbne.n\t%s\n", label);
  return 0;
}

static int asm_code_thumb_branch_cmp_eq(FILE *out, char *label, model_t model)
{
  ASM_CODE_OUT(out, "\tcmp\tr0, r1\n");
  ASM_CODE_OUT(out, "\tbeq.n\t%s\n", label);
  return 0;
}

static int asm_code_thumb_branch_cmp_ne(FILE *out, char *label, model_t model)
{
  ASM_CODE_OUT(out, "\tcmp\tr0, r1\n");
  ASM_CODE_OUT(out, "\tbne.n\t%s\n", label);
  return 0;
}

static int asm_code_thumb_branch_cmp_lt(FILE *out, char *label, model_t model)
{
  ASM_CODE_OUT(out, "\tcmp\tr0, r1\n");
  ASM_CODE_OUT(out, "\tblt.n\t%s\n", label);
  return 0;
}

static int asm_code_thumb_branch_cmp_gt(FILE *out, char *label, model_t model)
{
  ASM_CODE_OUT(out, "\tcmp\tr0, r1\n");
  ASM_CODE_OUT(out, "\tbgt.n\t%s\n", label);
  return 0;
}

static int asm_code_thumb_branch_cmp_le(FILE *out, char *label, model_t model)
{
  ASM_CODE_OUT(out, "\tcmp\tr0, r1\n");
  ASM_CODE_OUT(out, "\tble.n\t%s\n", label);
  return 0;
}

static int asm_code_thumb_branch_cmp_ge(FILE *out, char *label, model_t model)
{
  ASM_CODE_OUT(out, "\tcmp\tr0, r1\n");
  ASM_CODE_OUT(out, "\tbge.n\t%s\n", label);
  return 0;
}

static int asm_code_thumb_branch_cmp_ult(FILE *out, char *label, model_t model)
{
  ASM_CODE_OUT(out, "\tcmp\tr0, r1\n");
  ASM_CODE_OUT(out, "\tbcc.n\t%s\n", label);
  return 0;
}

static int asm_code_thumb_branch_cmp_ugt(FILE *out, char *label, model_t model)
{
  ASM_CODE_OUT(out, "\tcmp\tr0, r1\n");
  ASM_CODE_OUT(out, "\tbhi.n\t%s\n", label);
  return 0;
}

static int asm_code_thumb_branch_cmp_ule(FILE *out, char *label, model_t model)
{
  ASM_CODE_OUT(out, "\tcmp\tr0, r1\n");
  ASM_CODE_OUT(out, "\tbls.n\t%s\n", label);
  return 0;
}

static int asm_code_thumb_branch_cmp_uge(FILE *out, char *label, model_t model)
{
  ASM_CODE_OUT(out, "\tcmp\tr0, r1\n");
  ASM_CODE_OUT(out, "\tbcs.n\t%s\n", label);
  return 0;
}

static int asm_code_thumb_branch_cmp(FILE *out, char *label, c_type_t type, model_t model)
{
  return -1;
}

static void asm_code_thumb_function_call(FILE *out, char *label)
{
  ASM_CODE_OUT(out, "\tbl\t%s\n", label);
}

static void asm_code_thumb_function_call_pointer(FILE *out)
{
  asm_symbol_tmp_label_reserve(2);
  ASM_CODE_OUT(out, "\tbl\t%s\n", asm_symbol_tmp_label_namef(0));
  ASM_CODE_OUT(out, "\tb\t%s\n", asm_symbol_tmp_label_namef(1));
  asm_format_label(out, asm_symbol_tmp_label_name(0));
  ASM_CODE_OUT(out, "\tbx\tr0\n");
  asm_format_label(out, asm_symbol_tmp_label_name(1));
  asm_symbol_tmp_label_release(2);
}

static void asm_code_thumb_function_start(FILE *out)
{
  ASM_CODE_OUT(out, "\tpush\t{lr}\n");
}

static void asm_code_thumb_function_end(FILE *out)
{
  ASM_CODE_OUT(out, "\tpop\t{r1}\n");
  ASM_CODE_OUT(out, "\tbx\tr1\n");
}

static void asm_code_thumb_function_register_save(FILE *out)
{
  ASM_CODE_OUT(out, "\tpush\t{r4, r5}\n");
}

static void asm_code_thumb_function_register_load(FILE *out)
{
  ASM_CODE_OUT(out, "\tpop\t{r4, r5}\n");
}

struct asm_code_callback asm_code_callback_thumb = {
  asm_code_thumb_get_word_size,
  asm_code_thumb_get_pointer_size,

  asm_code_thumb_funccall_args_reg_number,
  asm_code_thumb_funccall_args_stack_number,
  asm_code_thumb_tmp_reg_number,
  asm_code_thumb_tmp_stack_number,
  asm_code_thumb_function_register_number,
  asm_code_thumb_function_saveparam_number,

  asm_code_thumb_stack_align_size,
  asm_code_thumb_stack_correct_size,

  asm_code_thumb_get_value,
  asm_code_thumb_get_value_r1,
  asm_code_thumb_get_address_stack,
  asm_code_thumb_get_address,
  asm_code_thumb_add_address,
  asm_code_thumb_get_r1,
  asm_code_thumb_set_r1,

  asm_code_thumb_memory_load,
  asm_code_thumb_memory_store,
  asm_code_thumb_stack_load,
  asm_code_thumb_stack_store,
  asm_code_thumb_stack_load_r1,
  asm_code_thumb_stack_store_r1,
  asm_code_thumb_stack_expand,
  asm_code_thumb_stack_reduce,

  asm_code_thumb_funccall_reg_load,
  asm_code_thumb_funccall_reg_store,
  asm_code_thumb_tmp_reg_load,
  asm_code_thumb_tmp_reg_save,
  asm_code_thumb_tmp_reg_load_r1,
  asm_code_thumb_tmp_reg_save_r1,

  asm_code_thumb_sign_extension_char,
  asm_code_thumb_sign_extension_uchar,
  asm_code_thumb_sign_extension_short,
  asm_code_thumb_sign_extension_ushort,
  asm_code_thumb_sign_extension_int,
  asm_code_thumb_sign_extension_uint,

  asm_code_thumb_calc_inv,
  asm_code_thumb_calc_minus,

  asm_code_thumb_calc_op1,

  asm_code_thumb_calc_add,
  asm_code_thumb_calc_sub,
  asm_code_thumb_calc_and,
  asm_code_thumb_calc_or,
  asm_code_thumb_calc_xor,

  asm_code_thumb_calc_mul,
  asm_code_thumb_calc_div,
  asm_code_thumb_calc_mod,
  asm_code_thumb_calc_llshift,
  asm_code_thumb_calc_rashift,
  asm_code_thumb_calc_rlshift,

  asm_code_thumb_calc_op2,

  asm_code_thumb_branch,
  asm_code_thumb_branch_zero,
  asm_code_thumb_branch_nzero,

  asm_code_thumb_branch_cmp_eq,
  asm_code_thumb_branch_cmp_ne,

  asm_code_thumb_branch_cmp_lt,
  asm_code_thumb_branch_cmp_gt,
  asm_code_thumb_branch_cmp_le,
  asm_code_thumb_branch_cmp_ge,
  asm_code_thumb_branch_cmp_ult,
  asm_code_thumb_branch_cmp_ugt,
  asm_code_thumb_branch_cmp_ule,
  asm_code_thumb_branch_cmp_uge,

  asm_code_thumb_branch_cmp,

  asm_code_thumb_function_call,
  asm_code_thumb_function_call_pointer,

  asm_code_thumb_function_start,
  asm_code_thumb_function_end,
  asm_code_thumb_function_register_save,
  asm_code_thumb_function_register_load,
};
